package CAIXS;
use Class::Accessor::Inherited::XS;
BEGIN { *CAIXS:: = *Class::Accessor::Inherited::XS:: }
1;
__END__

=head1 NAME

CAIXS - L<Class::Accessor::Inherited::XS> shortcut for less typing

=head1 DESCRIPTION

See L<Class::Accessor::Inherited::XS> documentation for more details.
